/* FILE: dspmon.c		Copyright (c), Symmetric Research, 1998-1999
 *
 * DSP32C monitor/debugger program.  This is the Win9x/WinNT version.
 * Because it automatically runs in a window, it does not use the /50 or
 * /tty command line options required for the DOS version.
 *
 *
 */

#include <stdio.h>
#include <conio.h>
#include <ctype.h>
#include <string.h>
#include <windows.h>

#include "dsptools.h"









/* LOW LEVEL SCREEN AND KEYBOARD SUPPORT:
 *
 * All low level interactions with the screen and keyboard are done
 * through the LL_ functions.
 *
 * In the functions given here, the Win 95/VC console function library
 * is used.  See the MS VC 6.0 documentation for details about console
 * apps.
 *
 * Previous versions of DSPMON have assumed they could write directly to
 * the frame buffer under DOS.  Using the console support functions is
 * more appropriate for Win 95 multitasking.
 *
 * DSPMON uses screen cell coordinates (i,j).  The upper left corner of
 * the screen is location (0,0), with i running down the screen and j
 * accross, (i,j) = (row,column).
 *
 *
 *
 *
 */


/* Screen functions ... */

#define SCREEN_WIDTH		80
#define SCREEN_HEIGHT		50

#define BLACK			0x00
#define BLUE			0x01
#define GREEN			0x02
#define CYAN			0x03
#define RED			0x04
#define MAGENTA			0x05
#define BROWN			0x06
#define GREY			0x07
#define INTENSE			0x08


#define COLOR( FORE, BACK)	(((BACK)<<4)|FORE)



HANDLE hOutput;


void LL_VideoSet( void) {

	COORD size, origin; long nbytes;

	    CONSOLE_CURSOR_INFO ConsoleCursorInfo;



	/* Get the screen output handle. */

	hOutput = GetStdHandle( STD_OUTPUT_HANDLE);




	/* Display the program title. */

	SetConsoleTitle( "SR DSPMON");




	/* Kick the screen into 80x50 mode. */
	
	size.X = 80;
	size.Y = 50;

	SetConsoleScreenBufferSize( hOutput, size);




	/* Clear the screen off. */

	origin.X = 0;
	origin.Y = 0;

	FillConsoleOutputCharacter( hOutput, ' ', 80*50, origin, &nbytes);




	/* Hide the console cursor. */

	ConsoleCursorInfo.dwSize   = 100;
	ConsoleCursorInfo.bVisible = FALSE;

	SetConsoleCursorInfo( hOutput, &ConsoleCursorInfo);



}
void LL_VideoRestore( void) {

	COORD size, origin; long nbytes; WORD atr;

	    CONSOLE_CURSOR_INFO ConsoleCursorInfo;

	        SMALL_RECT RestoreWindow;



	/* Clear the 80x50 screen off. */

	origin.X = 0;
	origin.Y = 0;

	atr = COLOR( GREY, BLACK);

	FillConsoleOutputCharacter( hOutput, ' ', 80*50, origin, &nbytes);
	FillConsoleOutputAttribute( hOutput, atr, 80*50, origin, &nbytes);
	


	/* Restore to standard 80x25. */

	size.X = 80;
	size.Y = 25;

	RestoreWindow.Left   = 0;
	RestoreWindow.Top    = 0;
	RestoreWindow.Right  = size.X - 1;
	RestoreWindow.Bottom = size.Y - 1;

	SetConsoleWindowInfo(       hOutput, TRUE, &RestoreWindow);
	SetConsoleScreenBufferSize( hOutput, size);




	/* Position the cursor at the origin. */

      	origin.X = 0;
      	origin.Y = 0;
      
      	SetConsoleCursorPosition( hOutput, origin);




	/* Redisplay the console cursor. */

	ConsoleCursorInfo.dwSize   = 12;
	ConsoleCursorInfo.bVisible = TRUE;

	SetConsoleCursorInfo( hOutput, &ConsoleCursorInfo);



}
void LL_PutC( int i, int j, int c, int color) {

	COORD loc; long nbytes;



	/* Display the character c at (i,j). */

	loc.X = j;
	loc.Y = i;

	FillConsoleOutputAttribute( hOutput, (WORD)color, 1, loc, &nbytes);
	FillConsoleOutputCharacter( hOutput, (TCHAR)c,    1, loc, &nbytes);



}
void LL_PutS( int i, int j, char *s, int n, int color) {

	COORD loc; int len; long nbytes;


	/* Display the string s at (i,j). */

	len = strlen( s);

        if ( len > n )

                len = n;



	loc.X = j;
	loc.Y = i;

	FillConsoleOutputAttribute( hOutput, (WORD)color, n, loc, &nbytes);

	WriteConsoleOutputCharacter( hOutput, s, len, loc, &nbytes);

	if ( len < n ) {

                /* fill out the balance of the line with spaces ... */

		loc.X += len;

		FillConsoleOutputCharacter( hOutput, ' ', n-len, loc, &nbytes);

		}


}

CHAR_INFO LL_ScreenSaveArea[SCREEN_HEIGHT*SCREEN_WIDTH];


void LL_ScreenSave( int i1, int j1, int ni, int nj) {

	COORD BufferSize, BufferCoord; SMALL_RECT ReadRegion;



	/* Save the rectangular area at (i1,j1) with size (ni,nj). */

	BufferSize.X      = SCREEN_WIDTH;
	BufferSize.Y      = SCREEN_HEIGHT;

	BufferCoord.X     = 0;
	BufferCoord.Y     = 0;

	ReadRegion.Left   = j1;
	ReadRegion.Top    = i1;
	ReadRegion.Right  = j1 + nj - 1;
	ReadRegion.Bottom = i1 + ni - 1;


	ReadConsoleOutput( hOutput,

				LL_ScreenSaveArea,
				BufferSize,
				BufferCoord,

				&ReadRegion

				);


}
void LL_ScreenRestore( int i1, int j1, int ni, int nj) {

	COORD BufferSize, BufferCoord; SMALL_RECT WriteRegion;



	/* Restore the rectangular area saved with ScreenSave. */

	BufferSize.X      = SCREEN_WIDTH;
	BufferSize.Y      = SCREEN_HEIGHT;

	BufferCoord.X     = 0;
	BufferCoord.Y     = 0;

	WriteRegion.Left   = j1;
	WriteRegion.Top    = i1;
	WriteRegion.Right  = j1 + nj - 1;
	WriteRegion.Bottom = i1 + ni - 1;


	WriteConsoleOutput( hOutput,

				LL_ScreenSaveArea,
				BufferSize,
				BufferCoord,

				&WriteRegion

				);
}





/* Keyboard functions ... */

#define F1		0x3B00		/* F1				*/
#define F2		0x3C00		/* F2				*/
#define F3		0x3D00		/* F3				*/
#define F4		0x3E00		/* F4				*/
#define F5		0x3F00		/* F5				*/
#define F6		0x4000		/* F6				*/
#define F7		0x4100		/* F7				*/
#define F8		0x4200		/* F8				*/
#define F9		0x4300		/* F9				*/
#define F10     	0x4400		/* F10  			*/

#define ESC		0x001B		/* esc				*/
#define ENTER		0x000D		/* enter			*/
#define BS		0x0008		/* backspace			*/
#define CTRL_BS		0x007F		/* ctrl backspace		*/
#define TAB		0x0009		/* tab				*/

#define HOME		0x4700		/* home				*/
#define END		0x4F00		/* end				*/
#define PGUP		0x4900		/* page up			*/
#define PGDN		0x5100		/* page down			*/

#define UP		0x4800		/* up arrow			*/
#define DOWN		0x5000		/* down arrow			*/
#define LEFT		0x4B00		/* left arrow			*/
#define RIGHT		0x4D00		/* right arrow			*/

#define CTRL_UP		0x8D00		/* ctrl up arrow                */
#define CTRL_DOWN	0x9100		/* ctrl down arrow              */
#define CTRL_LEFT	0x7300		/* ctrl left arrow		*/
#define CTRL_RIGHT	0x7400		/* ctrl right arrow		*/


#define FKEY_HELP	F1		/* function key assigments	*/
#define FKEY_SYMBOLS	F2
#define FKEY_QUIT	F3

#define FKEY_STEP	F5
#define FKEY_HALT	F6



int LL_KeyIn( void) {

	short c;



	/* Wait for a keypress, checking for function or arrow keys. */

	c = getch();

	if ( c == 0x0000 || c == 0x00E0 )

		c = getch() << 8;



	/* Return the keycode. */

	return( c);



}
int LL_KeyPress( void) {
	

	/* Return the keypress status, but don't wait. */

	if ( kbhit() )

	    { LL_KeyIn(); return( 1); }


	else
		return( 0);


}
















/* STRING HELPERS:
 *
 * Generic string helper functions.
 *
 *
 *
 *
 */

int StrEqu( char *src, char *dest) {


	/* Compare the src and dest for exact equality. */

	while ( *src && *dest && *src == *dest )

		{ src++; dest++; }


	return( *src == '\0' && *dest == '\0' ? 1 : 0 );


}
int StrScan( char *src, char *dest) {


	/* Scan src to see if it contains all of dest. */

	while ( *dest && *src == *dest )

		{ src++; dest++; }

	return( *dest == '\0' ? 1 : 0 );


}
char *StrChop( char *s) {

	char *p;


	/* Chop space off the end of a string and add \n. */

	p = s;

	while ( *p )

		p++;



	if ( *--p == '\n' ) {

		while ( *--p == ' ' )

			/* back up */ ;


		*++p = '\n';
		*++p = '\0';

		}


	return( s);

}














/* DOS COMMAND LINE USAGE:
 *
 *
 *
 *
 *
 */

char *DOS_UsageMsg[] = {

    "                                                                   \n",
    " dspmon [/n] [file.dso [file.dss]] [file.out]                      \n",
    "                                                                   \n",
    "      /n             cpu n, where n = 0,1,2,3  (DSP_MUL only)      \n",
    "                                                                   \n",
    "      file.dso       SR .dso object file to download               \n",
    "      file.dss       SR .dss symbol table file to use              \n",
    "      file.out       WE .out fully located coff file to download   \n",
    "                                                                   \n",
    "      where [] denotes an optional item                            \n",

    NULL,

    };


void DOS_Error( char *msg) {

	int i;



	/* Give the user an error message and quit. */

	printf( "ERROR: %s\n", msg);

	for ( i = 0 ; DOS_UsageMsg[i] ; i++ )

		printf( StrChop( DOS_UsageMsg[i]));


	exit( 1);


}
void DOS_Warning( char *msg) {


	/* Give the user a warning and wait for a key to continue on. */

	printf( "WARNING: %s, any key continues ... ", msg);

	    LL_KeyIn();


	printf( "\n");

}











/* DSPASM SYMBOL TABLE AND LOOKUP FUNCTIONS:
 *
 * These functions read in and manage a DSPASM .dss symbol table for DSPMON.
 *
 * NOTE: The defined constants DSPASM_XXXX *MUST* agree with DSPASM.
 *
 *
 *
 */

#define DSPASM_CONS		0x0001
#define DSPASM_BYTE		0x0002
#define DSPASM_WORD		0x0004
#define DSPASM_LONG		0x0008
#define DSPASM_DSPF		0x0010
#define DSPASM_IEEE		0x0020
#define DSPASM_FUNC		0x0040
#define DSPASM_GOTO		0x0080

#define DSPASM_DATA		0x003E      /* = BYTE|WORD|LONG|DSPF|IEEE  */
#define DSPASM_PROGRAM		0x00C0      /* = FUNC|GOTO                 */
#define DSPASM_ALL		0x00FF      /* = CONS|DATA|PROGRAM         */


typedef struct {

	unsigned long value;
	         char *name;
		   int type;
	}

	SYMBOL;





#define MAXSYMBOLS	300

SYMBOL symbol_table[ MAXSYMBOLS];

  unsigned nsymbols = 0;   /* actual number of symbols in table */




void SortSymbols( void) {

	int i, j; SYMBOL *s1, *s2, save;


	/* Sort the symbol table into value order. */

	for ( i = 1 ; i < nsymbols ; i++ ) {

		j = i-1;

		s1 = &symbol_table[i];
		s2 = &symbol_table[j];

		save = *s1;

		while ( j-- >= 0 && save.value < s2->value )

			*s1-- = *s2--;

		*s1 = save;

		}


}
SYMBOL *SymbolicValue( char *name, int *type) {

	unsigned i; SYMBOL *s;



	/* Do a linear search on the symbol table for the specified name. */

	for ( i = 0 , s = symbol_table ; i < nsymbols ; i++ , s++ ) {

		if ( StrEqu( name, s->name) )

			{ *type = s->type; return( s); }
		}

	return( NULL);



}
char *SymbolicName( unsigned long value, int type) {

	unsigned imin, imax, itest;

		unsigned long min, max, test;


	/* Binary search for a symbol of the specified value and type. */

	imin = 0;
	imax = nsymbols - 1;

	  min = symbol_table[imin].value;
	  max = symbol_table[imax].value;


	if ( nsymbols > 0 && min <= value && value <= max ) {

		while ( imin <= imax ) {

			itest = (imax + imin)/2;

			  test = symbol_table[itest].value;


			if ( value > test )

				imin = itest+1;

			else if ( value < test )

				imax = itest-1;

			else {

			     if ( symbol_table[itest].type & type )

				      return( symbol_table[itest].name);

				else
					return( "");

				}
			}
		}


	/* Failed return. */

	return( "");


}
int ReadSymbols( char *filename) {

	FILE *input;

	  SYMBOL *s;

	      char buffer[80], name[129]; unsigned long value;

		      int n, type;



	/* Open and read a DSPASM .dss ASCII symbol table file. */

	if ( !(input = fopen( filename, "r")) )

		return( 0);



	/* Check for the proper input file type. */

	if ( fscanf( input, "Global symbol values from %s", buffer) != 1 )

		return( 0);




	/* Read in the symbols. */

	       s = symbol_table;
	nsymbols = 0;


read:

      n = fscanf( input, "%128s %lx %s (%x)\n", name, &value, buffer, &type);

	  if ( n == 4 ) {

		s->name  = strdup( name);
		s->value = value;
		s->type  = type;

		  if ( s->name && nsymbols < MAXSYMBOLS )

			  { s++; nsymbols++; goto read; }
		}





	/* Sort the table into value order. */

	SortSymbols();




	/* Check to see if the entire file was read or not. */

	if ( feof( input) )

		return( 1);


	else
                /* return success if at least some symbols were read ... */

                return( nsymbols > 0 ? 1 : 0 );



}
void ClearSymbolTable( void) {

	/* Clear the symbol table. */

	nsymbols = 0;

}









/* DSP32C EXECUTION AND BREAKPOINT CONTROL:
 *
 * The following is for controlling the execution of the DSP32C.  Both
 * the g and s DSPMON commands use this code.  Note that is the DSP32C
 * never reaches a breakpoint, then any PC keyboard hit will interrupt
 * the execution and return control to the monitor.
 *
 *
 * NOTES:
 *
 * * ALL BREAKPOINTING IS DONE IN SOFTWARE
 *
 * The DSP32C does not have hardware breakpointing, so breakpointing must
 * be handled in software.  The scheme is as follows:
 *
 * First, breakpoint code is downloaded to the DSP32C.  Then a goto is
 * planted in the user's code at the breakpoint address.  When execution
 * hits the planted goto and jumps to the breakpoint code, the DSP32C
 * registers are rolled out and the PC is notified that the breakpoint
 * has occurred.
 *
 * After notifying the PC of the break, the DSP32C goes into an infinite
 * loop waiting for a signal from the PC that is should continue on.  To
 * signal the DSP32C to restart, the PC pokes the breakpoint code with a
 * goto instruction returning to the user's code.  Basically, this is
 * self modifying code, and seems to be the only way to have breakpoints
 * without modifying the CAU and DAU flags.
 * 
 * Note that the breakpoint code trashes the value of register r21.  All
 * other CAU and DAU registers and flags are preserved.
 * 
 *
 * * THE BREAKPOINT CODE DSPMON.DSH FROM DSPMON.DSP IS INCLUDED
 *
 * The DSP32C code dspmon.dsp is assembled with the DSPASM /h option to
 * create a hexified code array dspmon.dsh that is included in this file.
 * Including the code as an array in dspmon.exe eliminates the need for
 * having an additional dspmon.dso file around that must be found each
 * time dspmon.exe is started.
 *
 *
 * * THE SYMBOL TABLE DSPMON.DSC FROM DSPMON.DSP IS INCLUDED
 *
 * The DSP32C code dspmon.dsp is also assembled with the DSPASM /sc
 * option to create a symbol table dspmon.dsc for inclusion in this
 * file.  Symbols which need to be known in both dspmon.dsp and dspmon.c
 * are automatically communicated to dspmon.c by this technique.  Note
 * that all dspmon.dsp symbols are automatically given a DSP_ prefix by
 * DSPASM in the C include file dspmon.dsc.
 *
 *
 * * DON'T OVERWRITE THE BREAKPOINT CODE ON THE DSP32C
 *
 * Currently, the DSP32C breakpoint code is downloaded to the internal
 * ram block RAM1 in the DSP32C high memory.  This memory is always present
 * at the same addresses for all DSP32C memory mode configurations,
 * and is probably the best choice for most users.  The base address to
 * download the breakpoint code to is given by the defined constant
 * BREAK_BASE in dspmon.dsp.
 *
 * When the breakpoint code is downloaded, it will overwrite any other
 * program or data there.  User applications should not have program or
 * data in ram block RAM1, and likewise should be careful not to
 * overwrite those areas if breakpointing is to work correctly.
 *
 *
 *
 */


/* Include the hexified code and symbol table from dspmon.dsp. */

#include "dspmon.dsh"
#include "dspmon.dsc"




/* DSP32C register stack. */

struct {

	/* This *MUST* agree with dspmon.dsp. */

        unsigned long cau[0x16];

	  union { unsigned long l; float f; } dau[0x4];

	     struct cau_flags { unsigned long n, z, v, c; } cau_flags;
	     struct dau_flags { unsigned long n, z, v, u; } dau_flags;

	}

	regs;





/* DSP32C instructions. */

#define H_MASK		0x00FF0000L
#define L_MASK		0x0000FFFFL

#define R1_equ( n)	0xC0010000L | ((n&H_MASK)<<5) | (n&L_MASK)
#define Goto( a)	0xA0000000L | ((a&H_MASK)<<5) | (a&L_MASK)
#define Nop		0x00000000L

unsigned long

    users_inst[3],

	break_inst[3] = {  /* instructions to plant at the breakpoint */
	
		Nop,
		Goto( DSP_break),
		Nop

		},

	  restart[2];





/* Current execution status and breakpoint address. */

#define HALTED		1	/* execution status values */
#define EXECUTING	2
#define BREAKPOINTING	3

  unsigned execution_status = HALTED; unsigned long breakpoint = 0;




void SetBreakpoint( void) {

	unsigned long signal;



	/* Save the user's instructions at the breakpoint. */
	
	DSP32C_Get( breakpoint, users_inst, sizeof( users_inst));



	/* Plant a breakpoint branch in the user's code. */

	DSP32C_Put( break_inst, breakpoint, sizeof( break_inst));



	/* Clear the DSP32C's break signal. */

	signal = 0;

	    DSP32C_Put( &signal, DSP_signal, sizeof( signal));



}
void RestartFromBreakpoint( void) {


	/* Restart from the DSP32C's breakpoint code. */

	DSP32C_PutLong( Goto( DSP_break_restart), DSP_break_wait);



}
int WaitForBreak( void) {

	unsigned long signal; int status;

	    void PopupClock( unsigned nterminal);



	/* Wait for the DSP32C's breakpoint signal or a kb interrupt. */

	while ( 1 ) {

	  PopupClock( 0x800); /* show the DSPMON busy clock */

	    DSP32C_Get( DSP_signal, &signal, sizeof( signal));



		/* If the break occured, get the register dump. */

		if ( signal ) {
			
			DSP32C_Get( DSP_regs, &regs, sizeof( regs));

			  status = 1; break;

			}



		/* Otherwise, check for a kb interrupt. */

		else if ( LL_KeyPress() )
			
			    { DSP32C_Halt(); status = 0; break; }

	    }



	/* Clear the DSPMON busy clock. */

	PopupClock( 0);



	/* Restore the user's instructions at the breakpoint. */

	DSP32C_Put( users_inst, breakpoint, sizeof( users_inst));



	/* Poke the DSP32C with restart code for the next breakpoint. */

	restart[0] = R1_equ( regs.cau[0]);
	restart[1] =   Goto( breakpoint);

	  DSP32C_Put( restart, DSP_break_restart, sizeof( restart));



	/* Return whether the breakpoint occured or got a kbhit. */

	return( status);



}
int ExecuteBreakpoint( void) {

	unsigned status;
	


	/* If this is the first breakpoint, download and start. */
	
	if ( execution_status == HALTED ) {
		
		DSP32C_SrHexLoad( dspmon_dsh);
		
		    SetBreakpoint();

		    DSP32C_Go();

		}

	
	/* Otherwise restart from a previous breakpoint. */
	
	else /* execution_status == BREAKPOINTING */ {
		
		SetBreakpoint();

		    RestartFromBreakpoint();

		}



	/* Wait for the breakpoint. */
	
	return( WaitForBreak() );

}








/* DSPMON CHARACTER WINDOWS:
 *
 * These helpers are built on top of the LL_ low level functions.
 *
 *
 *
 */

typedef struct window {

	char *title;

	int
		ititle,
		jtitle,
		ctitle_on,
		ctitle_off,

		iscroll,
		jscroll,
		hscroll,
		wscroll,
		cscroll;

	void
		  (*up)( void),
		(*down)( void),
		(*pgup)( void),
		(*pgdn)( void),
		(*home)( void),
		 (*end)( void);
	}

	WINDOW;



/* Forward declarations of window structure functions. */

void

        DataDisplayUp( void),
        DataDisplayDown( void),
        DataDisplayPgUp( void),
        DataDisplayPgDn( void),
        DataDisplayHome( void),
        DataDisplayEnd( void),

        UnassembleUp( void),
        UnassembleDown( void),
        UnassemblePgup( void),
        UnassemblePgdn( void),
        UnassembleHome( void),
        UnassembleEnd( void);




/* DSPMON memory display window. */

#define D_I			0
#define D_J			0
#define D_N			5

#define D_ITITLE		D_I
#define D_JTITLE		D_J
#define D_CTITLE_ON		COLOR( INTENSE|GREY, BLACK)
#define D_CTITLE_OFF		COLOR( GREY, BLACK)

#define D_ISCROLL		D_I+1
#define D_JSCROLL		D_J
#define D_HSCROLL		D_N
#define D_WSCROLL		SCREEN_WIDTH
#define D_CSCROLL		COLOR( GREY, INTENSE|BLACK)


/* DSPMON unassembly window. */

#define U_I			6
#define U_J			0
#define U_N			31

#define U_ITITLE		U_I
#define U_JTITLE		U_J
#define U_CTITLE_ON		COLOR( INTENSE|GREY, BLACK)
#define U_CTITLE_OFF		COLOR( GREY, BLACK)

#define U_ISCROLL		U_I+1
#define U_JSCROLL		U_J
#define U_HSCROLL		U_N
#define U_WSCROLL		SCREEN_WIDTH
#define U_CSCROLL		COLOR( INTENSE|CYAN, CYAN)

#define U_CBREAKPOINT		COLOR( INTENSE|GREY, GREY)



/* DSPMON register window. */

#define R_I			38
#define R_J			0
#define R_N			10

#define R_ITITLE		R_I
#define R_JTITLE		R_J
#define R_CTITLE_ON		COLOR( INTENSE|GREY, BLACK)
#define R_CTITLE_OFF		COLOR( GREY, BLACK)

#define R_ISCROLL		R_I+1
#define R_JSCROLL		R_J
#define R_HSCROLL		R_N
#define R_WSCROLL		SCREEN_WIDTH
#define R_CSCROLL		COLOR( GREY, INTENSE|BLACK)


/* Help popup window. */

#define H_I			5
#define H_J			10
#define H_N			40

#define H_ITITLE		H_I
#define H_JTITLE		H_J
#define H_CTITLE_ON		COLOR( INTENSE|GREY, BLACK)
#define H_CTITLE_OFF		COLOR( GREY, BLACK)

#define H_ISCROLL		H_I+1
#define H_JSCROLL		H_J
#define H_HSCROLL		H_N
#define H_WSCROLL		SCREEN_WIDTH-2*H_J
#define H_CSCROLL		COLOR( INTENSE|GREY, GREY)



/* Error message popup window. */

#define E_I			41
#define E_J			5
#define E_N			3

#define E_ITITLE		E_I
#define E_JTITLE		E_J
#define E_CTITLE_ON		COLOR( INTENSE|GREY, GREY)
#define E_CTITLE_OFF		COLOR( GREY, BLACK)

#define E_ISCROLL		E_I+1
#define E_JSCROLL		E_J
#define E_HSCROLL		E_N
#define E_WSCROLL		SCREEN_WIDTH-2*E_J
#define E_CSCROLL		COLOR( INTENSE|RED, RED)



/* Message popup window. */

#define G_I			41
#define G_J			5
#define G_N			3

#define G_ITITLE		G_I
#define G_JTITLE		G_J
#define G_CTITLE_ON		COLOR( INTENSE|GREY, GREY)
#define G_CTITLE_OFF		COLOR( GREY, BLACK)

#define G_ISCROLL		G_I+1
#define G_JSCROLL		G_J
#define G_HSCROLL		G_N
#define G_WSCROLL		SCREEN_WIDTH-2*G_J
#define G_CSCROLL		COLOR( INTENSE|MAGENTA, MAGENTA)



/* Symbol table popup window. */

#define S_I			4
#define S_J			8
#define S_N			40

#define S_ITITLE		S_I
#define S_JTITLE		S_J
#define S_CTITLE_ON		COLOR( INTENSE|GREY, BLACK)
#define S_CTITLE_OFF		COLOR( GREY, BLACK)

#define S_ISCROLL		S_I+1
#define S_JSCROLL		S_J
#define S_HSCROLL		S_N
#define S_WSCROLL		SCREEN_WIDTH-2*S_J
#define S_CSCROLL		COLOR( INTENSE|MAGENTA, MAGENTA)




/* Static window allocation. */

WINDOW

    display_memory = {

	"Display memory",

	    D_ITITLE, D_JTITLE, D_CTITLE_ON, D_CTITLE_OFF,

	      D_ISCROLL, D_JSCROLL, D_HSCROLL, D_WSCROLL, D_CSCROLL,

	    DataDisplayUp,
	    DataDisplayDown,
	    DataDisplayPgUp,
	    DataDisplayPgDn,
	    DataDisplayHome,
	    DataDisplayEnd,

	    },

    unassemble_memory = {

	"Unassemble memory",

	    U_ITITLE, U_JTITLE, U_CTITLE_ON, U_CTITLE_OFF,

	      U_ISCROLL, U_JSCROLL, U_HSCROLL, U_WSCROLL, U_CSCROLL,

	    UnassembleUp,
	    UnassembleDown,
	    UnassemblePgup,
	    UnassemblePgdn,
	    UnassembleHome,
	    UnassembleEnd,

	    },


    register_display = {

	"Registers",

	    R_ITITLE, R_JTITLE, R_CTITLE_ON, R_CTITLE_OFF,

	      R_ISCROLL, R_JSCROLL, R_HSCROLL, R_WSCROLL, R_CSCROLL,

	    },


    help_window = {

	"Help (any key continues) ...",

	    H_ITITLE, H_JTITLE, H_CTITLE_ON, H_CTITLE_OFF,

	      H_ISCROLL, H_JSCROLL, H_HSCROLL, H_WSCROLL, H_CSCROLL,

	    },


    error_window = {

	"Error (any key continues, F1 = help)",

	    E_ITITLE, E_JTITLE, E_CTITLE_ON, E_CTITLE_OFF,

	      E_ISCROLL, E_JSCROLL, E_HSCROLL, E_WSCROLL, E_CSCROLL,

	    },


    msg_window = {

	"Message (any key continues, F1 = help)",

	    G_ITITLE, G_JTITLE, G_CTITLE_ON, G_CTITLE_OFF,

	      G_ISCROLL, G_JSCROLL, G_HSCROLL, G_WSCROLL, G_CSCROLL,

	    },


    symbol_window = {

	"Symbol table (arrow keys scroll, other keys continue) ...",

	    S_ITITLE, S_JTITLE, S_CTITLE_ON, S_CTITLE_OFF,

	      S_ISCROLL, S_JSCROLL, S_HSCROLL, S_WSCROLL, S_CSCROLL,

	    };




#define ON	1
#define OFF	0



WINDOW *whighlighted = NULL;	/* current active window */


void WSave( WINDOW *w) {

	int i1, j1, ni, nj;


	/* Save the area underlying w. */

	i1 = w->ititle;
	j1 = w->jtitle;

	ni = w->hscroll + 1;
	nj = w->wscroll;
	

	LL_ScreenSave( i1, j1, ni, nj);



}
void WRestore( WINDOW *w) {

	int i1, j1, ni, nj;



	/* Restore the area underlying w. */

	i1 = w->ititle;
	j1 = w->jtitle;

	ni = w->hscroll + 1;
	nj = w->wscroll;

	LL_ScreenRestore( i1, j1, ni, nj);



}
void WTitle( WINDOW *w, int state) {

	int i1, j1, nj, c;


	/* Display a window title in the specified state. */

	i1 = w->ititle;
	j1 = w->jtitle;
	nj = w->wscroll;

	c = ( state == ON ? w->ctitle_on : w->ctitle_off );

	    LL_PutS( i1, j1, w->title, nj, c);


}
void WSelect( WINDOW *wnew) {


	/* Turn the title hightlight off on the current window. */

	if ( whighlighted )

		WTitle( whighlighted, OFF);



	/* Set the current window to the new one, and highlight it. */

	WTitle( wnew, ON);

	    whighlighted = wnew;




}
void PopupClock( unsigned nterminal) {

	static unsigned first_call = 1;

	    static unsigned tick, c;

#define ICLOCK	49
#define JCLOCK	79



	/* Display the clock if the first call. */

	if ( first_call ) {

		first_call = 0; tick = 0;

		  c = COLOR( INTENSE|GREEN, BLACK);

		}



	/* If tick equals the terminal count, toggle the clock mark. */

	if ( tick++ == nterminal ) {

		LL_PutC( ICLOCK, JCLOCK, '', c);

		    tick = 0; c++; if ( c == 0x10 ) c = 1;

		}



	/* Reinitialize and clear the clock mark if nterminal = 0. */

	if ( nterminal == 0 ) {

		LL_PutC( ICLOCK, JCLOCK, ' ', COLOR( BLACK, BLACK));

		    first_call = 1;

		}



}
void PopupError( char *msg, int hotkey) {

	int i1, j1, nj, color; char buffer[SCREEN_WIDTH];

#define BLANKL		""



	/* Save the area underlying the error popup. */

	WSave( &error_window);




	/* Popup the error message. */

	WTitle( &error_window, ON);

	   i1 = error_window.iscroll;
	   j1 = error_window.jscroll;
	   nj = error_window.wscroll;
	color = error_window.cscroll;

	sprintf( buffer, "    %s ...", msg);


	  LL_PutS( i1++, j1, BLANKL, nj, color);
	  LL_PutS( i1++, j1, buffer, nj, color);
	  LL_PutS( i1++, j1, BLANKL, nj, color);




	/* Wait for a keypress other than the hot key to continue. */

	while ( LL_KeyIn() == hotkey )

		;



	/* Restore the area underlying the popup. */

	WRestore( &error_window);



	/**** CLEAR OFF THE COMMAND LINE ????? ****/



}
void PopupMsg( char *msg, int hotkey) {

	int i1, j1, nj, color; char buffer[SCREEN_WIDTH];




	/* Save the area underlying the popup message. */

	WSave( &msg_window);



	/* Format the message. */

	sprintf( buffer, "    %s ...", msg);



	/* Popup the message window. */

	   i1 = msg_window.iscroll;
	   j1 = msg_window.jscroll;
	   nj = msg_window.wscroll;
	color = msg_window.cscroll;

	WTitle( &msg_window, ON);

	  LL_PutS( i1++, j1, BLANKL, nj, color);
	  LL_PutS( i1++, j1, buffer, nj, color);
	  LL_PutS( i1++, j1, BLANKL, nj, color);




	/* Wait for a keypress other than the hot key to continue. */

	while ( LL_KeyIn() == hotkey )

		;



	/* Restore the area underlying the popup. */

	WRestore( &msg_window);



	/*** CLEAR OFF THE COMMAND LINE? ***/



}

char *PopupHelpMsg[] = {

	"                                                           ",
	"   Commands                                            ",
	"                                                           ",
	"   d[type] [addr]        display memory                    ",
	"   e[type] [addr]        enter memory value                ",
	"                                                           ",
	"   c [addr] [nbytes]     clear memory                      ",
	"                                                           ",
	"   u [addr]              unassemble memory                 ",
	"                                                           ",
	"   l filename            load a dso, dss, or out file      ",
	"                                                           ",
	"   ?             <F1>    help                              ",
	"   symbols       <F2>    display symbol table              ",
	"   eni                   enable eni interrupt flag         ",
	"                                                           ",
	"   q             <F3>    quit                              ",
	"                                                           ",
	"   ESC                   clear command buffer              ",
	"                                                           ",
	"   g [addr]              go with optional breakpoint       ",
	"   s             <F5>    single step                       ",
	"   h             <F6>    halt and reset pc to 0            ",
	"                                                           ",
	"                                                           ",
	"    where type = (b)yte, (w)ord, (l)ong, dsp(f), (i)eee    ",
	"      addresses may be symbolic if dss file is loaded      ",
	"                                                           ",
	"                                                           ",
	"   Arrow Keys                                          ",
	"                                                           ",
	"          tab, up, down, home, end, pgup, pgdn             ",
	"      ctrl left, ctrl right, ctrl pgup, ctrl pgdn          ",
	"                                                           ",
	"                                                           ",

	NULL,

	};


void PopupHelp( void) {

        int i1, j1, ni, nj, i2, j2, i, color, imsg;



	/* Save the area underlying the help popup. */

	WSave( &help_window);



	/* Popup a help window. */

	WTitle( &help_window, ON);


	   i1 = help_window.iscroll;
	   j1 = help_window.jscroll;
	   ni = help_window.hscroll;
	   nj = help_window.wscroll;
        color = help_window.cscroll;

	   i2 = i1 + ni - 1;

	for ( i = i1 , imsg = 0 ; i <= i2 && PopupHelpMsg[imsg] ; i++ )

		LL_PutS( i, j1, PopupHelpMsg[imsg++], nj, color);




	/* Wait for a keypress other than the hot key to continue. */

	while ( LL_KeyIn() == FKEY_HELP )
		
		;



	/* Restore the area underlying the help popup. */

	WRestore( &help_window);



}
void PopupSymbols( void) {

	int i1, j1, ni, nj, i2, j2, color, i, npage;


	  SYMBOL *start, *s; char buffer[SCREEN_WIDTH];





	/* Special case no symbols. */

	if ( nsymbols == 0 )

		{ PopupError( "No symbol table loaded", 0); return; }




	/* Save the screen area underlying the symbol table popup. */

	WSave( &symbol_window);





	/* Display one page of the symbol table. */

	WTitle( &symbol_window, ON);

	   i1 = symbol_window.iscroll;
	   j1 = symbol_window.jscroll;
	   ni = symbol_window.hscroll;
	   nj = symbol_window.wscroll;
	color = symbol_window.cscroll;

	npage = ni - 2;

	start = symbol_table;


  display_page:

	  s = start;

	  i = i1;

	    LL_PutS( i++, j1, "", nj, color);

	      while ( i < (i1+npage) ) {

		  if ( s < symbol_table+nsymbols ) {

			  sprintf( buffer, "      0x%08lX:   %-31.31s",

							  s->value,
							  s->name
						);

			      LL_PutS( i++, j1, buffer, nj, color);

			  s++;

			  }

		  else
			  LL_PutS( i++, j1, "", nj, color);

		  }

	LL_PutS( i++, j1, "", nj, color);






	/* Wait for a keypress, paging if requested. */

  get_page:

	switch ( LL_KeyIn() ) {

		case HOME:

			start = symbol_table;
			goto display_page;

		case END:

			start = symbol_table+nsymbols-npage+1;

			if ( start < symbol_table )

				start = symbol_table;

			goto display_page;

		case UP:

			start--;

			if ( start < symbol_table )

				start = symbol_table;

			goto display_page;

		case DOWN:

			start++;

			if ( start > symbol_table+nsymbols )

				start--;

			goto display_page;


		case PGUP:

			start -= npage;

			if ( start < symbol_table )

				start = symbol_table;

			goto display_page;

		case PGDN:

			start += npage;

			if ( start > symbol_table+nsymbols )

				start -= npage;

			goto display_page;


		case FKEY_SYMBOLS:

		  /* require something other than the hot key to continue */

			goto get_page;

		}



	/* Restore the popup area. */

	WRestore( &symbol_window);



}






/* COMMAND LINE BUFFER AND POINTER:
 *
 * Responses the user types in at the command line are stored here.
 *
 *
 *
 */

#define MAXCOMMAND	50

char  command_buffer[MAXCOMMAND];
char *cbuf;
int   command_length = 0;












void UpdateWindows( void);




/* DISPLAY MEMORY COMMAND:
 *
 *
 *
 *
 */

struct display_data {

        int type;
        unsigned long start;

        }

        d = { DSPASM_LONG, 0L };




char *dprintf( int type, unsigned long addr) {

	int b; static char buf[80]; char *name;

	union {

		unsigned char  b[4];
		unsigned short w[2];
		unsigned long  l;
		float          f;
		float          i;

		} data;



	/* Read and format DSP32C data. */

#define PRINTABLE( c)	(isprint( c) ? c : '.')


	DSP32C_Get( addr, &data, sizeof( data));


	name = SymbolicName( addr, DSPASM_DATA);

	  b = sprintf( &buf[0], "%12.12s %06lX: ", name, addr&0xFFFFFFL);


	switch ( type ) {

	    case DSPASM_BYTE:

		b += sprintf( &buf[b],

			"%02X %02X %02X %02X      ASCII [%c%c%c%c]",

				data.b[0], data.b[1], data.b[2], data.b[3],

					PRINTABLE( data.b[0]),
					PRINTABLE( data.b[1]),
					PRINTABLE( data.b[2]),
					PRINTABLE( data.b[3])
				);
		break;


	    case DSPASM_WORD:

		b += sprintf( &buf[b], "%04X %04X", data.w[0], data.w[1]);
		break;


	    case DSPASM_LONG:

		b += sprintf( &buf[b], "%08lX", data.l);
		break;


	    case DSPASM_DSPF:

		b += sprintf( &buf[b], "%s    DSPF", DSP32C_DspfToA( &data.f));
		break;


	    case DSPASM_IEEE:

		b += sprintf( &buf[b], "%s    IEEE", DSP32C_IeeeToA( &data.i));
		break;

		}


	return( buf);


}
void DataDisplay( void) {

	int i1, j1, ni, nj, i2, i; unsigned long addr;



	/* Display the window title bar. */

	if ( whighlighted == &display_memory )

		WTitle( &display_memory, ON);

	else

		WTitle( &display_memory, OFF);




	/* Display DSP32C memory. */

	i1 = display_memory.iscroll;
	j1 = display_memory.jscroll;
	ni = display_memory.hscroll;
	nj = display_memory.wscroll;

	i2 = i1 + ni - 1;


	addr = d.start;

	for ( i = i1 ; i <= i2 ; i++ , addr += 4 )

		LL_PutS( i, j1, dprintf( d.type, addr), nj, D_CSCROLL);



}
void DataDisplayUp( void)

	{ d.start -= 4; DataDisplay(); }

void DataDisplayDown( void)

	{ d.start += 4; DataDisplay(); }

void DataDisplayPgUp( void)

	{ d.start -= display_memory.hscroll*4; DataDisplay(); }

void DataDisplayPgDn( void)

	{ d.start += display_memory.hscroll*4; DataDisplay(); }

void DataDisplayHome( void)

	{ d.start = 0L; DataDisplay(); }

void DataDisplayEnd( void)

	{ d.start = 0L - display_memory.hscroll*4; DataDisplay(); }








/* ENTER MEMORY VALUE COMMAND:
 *
 *
 *
 */

struct

    enter_data { int type; unsigned long addr; } e = { DSPASM_LONG, 0L };


void enter( void) {

	unsigned long addr;

	   char *type, msg[MAXCOMMAND]; unsigned n;


	union {

		unsigned short w;
		unsigned long  l;
		float          f;
		float          i;

		} value;


	int
	     space( void),

		hexnum( int ndigits, unsigned long *value),

		    eob( void);


	void CommandGets( char *msg);





	/* Don't allow byte entry, the PIO is 16 bits wide ... */

	if ( e.type == DSPASM_BYTE ) {

		PopupError( "Byte entry not allowed, try word etc", 0);
		
		return;

		}



	/* Let the user enter some new values. */

	switch ( e.type ) {

		case DSPASM_WORD: type = "WORD"; break;
		case DSPASM_LONG: type = "LONG"; break;
		case DSPASM_DSPF: type = "DSPF"; break;
		case DSPASM_IEEE: type = "IEEE"; break;

		}

	addr = e.addr;



  get_value:

	sprintf( msg, "Value (%s) for 0x%06lX ? ", type, addr&0xFFFFFFL);

	   CommandGets( msg); space();



	     /* Poke the DSP32C with the value. */

	     switch ( e.type ) {

		case DSPASM_WORD:

			if ( hexnum( 4, &value.l) && eob() ) {

			    DSP32C_Put( &value.w, addr, sizeof( unsigned));

				  addr += 2; break;
				}

			return;


		case DSPASM_LONG:

			if ( hexnum( 8, &value.l) && eob() ) {

			    DSP32C_Put( &value.l, addr, sizeof( unsigned long));

				  addr += 4; break;
				}

			return;


		case DSPASM_DSPF:

			n = DSP32C_AToDspf( cbuf, &value.f);

			if ( n ) {

			  cbuf += n; if ( eob() ) {

			    DSP32C_Put( &value.l, addr, sizeof( unsigned long));

				  addr += 4; break;
			       }
		           }

			return;


		case DSPASM_IEEE:

			n = DSP32C_AToIeee( cbuf, &value.i);

			if ( n ) {

			  cbuf += n; if ( eob() ) {

			    DSP32C_Put( &value.l, addr, sizeof( unsigned long));

				  addr += 4; break;
				}
			    }

			return;

		}



		/* Update the display, and get the next value. */

		UpdateWindows();


		goto get_value;
}





/* CLEAR MEMORY COMMAND:
 *
 *
 *
 *
 */

struct

    clear_data { unsigned long start, nbytes; } c = { 0L, 0x80L };


void clear( void) {

	unsigned long addr;

	  long nwords;

	    static unsigned long zero = 0L;



	/* Special case nbytes = 0. */

	if ( c.nbytes == 0 )

		return;



	/* Clear off memory, rounding nbytes up to words. */

	  addr = c.start;
	nwords = (c.nbytes+1)/2;


	while ( nwords-- > 0 ) {


	    PopupClock( 0x800); /* show the busy clock */


		/* Clear off a location. */

		DSP32C_Put( &zero, addr, sizeof( unsigned short)); addr += 2;



		/* Let the user break out with a keypress hit if needed. */

		if ( LL_KeyPress() ) {

		    PopupError( "Keyboard interrupt, clear not finished", 0);

		         break;

		     }

		}



	/* Reset and clear the busy clock. */

	PopupClock( 0);



	/* Update the display. */

	UpdateWindows();


}





/* UNASSEMBLE (DIASSEMBLE) MEMORY COMMAND:
 *
 *
 *
 *
 *
 */

struct

    unassemble_data { unsigned long start; } u = { 0L };


char *uprintf( unsigned long addr) {

	unsigned long value;

	  char inst[SCREEN_WIDTH], *name; static char buf[SCREEN_WIDTH];


	/* Disassemble a DSP32C memory location. */

	DSP32C_Get( addr, &value, sizeof( unsigned long));

	  DSP32C_Decode( value, inst);

	    name = SymbolicName( addr, DSPASM_PROGRAM);

	      sprintf( buf, "%12.12s %06lX: %08lX  %-48.48s",

					name,
					addr & 0x00FFFFFFL,
					value,
					inst

					);
	  return( buf);


}
void unassemble( void) {

	int i1, j1, ni, nj, i2, i; unsigned long addr; 



	/* Display the window title bar. */

	if ( whighlighted == &unassemble_memory )

		WTitle( &unassemble_memory, ON);

	else
		WTitle( &unassemble_memory, OFF);




	/* Display DSP32C memory as unassembled mnemonics. */

	i1 = unassemble_memory.iscroll;
	j1 = unassemble_memory.jscroll;
	ni = unassemble_memory.hscroll;
	nj = unassemble_memory.wscroll;

	i2 = i1+ni-1;


	addr = u.start;

	for ( i = i1 ; i <= i2 ; i++ , addr += 4 ) {


		/* Put a highlight bar on the breakpoint address. */

		if ( addr == breakpoint )

			LL_PutS( i, j1, uprintf( addr), nj, U_CBREAKPOINT);

		else
			LL_PutS( i, j1, uprintf( addr), nj, U_CSCROLL);


	      }

}
void UnassembleUp( void)

	{ u.start -= 4; unassemble(); }

void UnassembleDown( void)

	{ u.start += 4; unassemble(); }

void UnassemblePgup( void)

	{ u.start -= unassemble_memory.hscroll*4; unassemble(); }

void UnassemblePgdn( void)

	{ u.start += unassemble_memory.hscroll*4; unassemble(); }

void UnassembleHome( void)

	{ u.start = 0L; unassemble(); }

void UnassembleEnd( void)

	{ u.start = 0L - unassemble_memory.hscroll*4; unassemble(); }






/* LOAD AN INPUT FILE COMMAND:
 *
 *
 *
 *
 */

void load( char *filename) {


	/* Try to download a SR .dso object file.
	 *
	 * Note that SrLoad automatically does a halt when downloading,
	 * and the breakpoint and execution status are updated to reflect
	 * that.
	 *
	 *
	 *
	 */

	if ( DSP32C_SrLoad( filename, NO_REPORT) ) {

		      breakpoint = 0;
		execution_status = HALTED;

		ClearSymbolTable();

		         u.start = 0;

		}



	/* Try to load a SR .dss symbol table file. */

	else if ( ReadSymbols( filename) )

		/* downloaded a .dss symbol table file */ ;



	/* Try to download a WE AT&T .out coff object file. */

	else if ( DSP32C_WeLoad( filename, NO_REPORT) ) {

		breakpoint = 0; execution_status = HALTED;

		  ClearSymbolTable();

		    u.start = 0;
		}


	/* Unrecognized file format. */

	else {

		PopupError( "Unable to read input file", 0);

		return;

		}



	/* Update the display. */

	UpdateWindows();

}





/* REGISTER DISPLAY:
 *
 *
 *
 *
 *
 *
 */

struct

    register_data { WINDOW *w; } r = { &register_display };


char          /* flag and register display formats */

    *flag_fmt = "%-46s CAU(%c%c%c%c) DAU(%c%c%c%c) (0x%06lX)",

       *unknown_flag_fmt = "%-46s CAU(****) DAU(****) (0x%06lX)",


 *register_fmt[] = {

   "         r1 0x%06lX    r7 0x%06lX   r13 0x%06lX   r19 0x%06lX         ",
   "         r2 0x%06lX    r8 0x%06lX   r14 0x%06lX   r20 0x%06lX         ",
   "         r3 0x%06lX    r9 0x%06lX   r15 0x%06lX   r21 0x%06lX  (****) ",
   "         r4 0x%06lX   r10 0x%06lX   r16 0x%06lX   r22 0x%06lX  (ivtp) ",
   "         r5 0x%06lX   r11 0x%06lX   r17 0x%06lX                       ",
   "         r6 0x%06lX   r12 0x%06lX   r18 0x%06lX                       ",
   "                                                                      ",
   "      a0 %-13s (0x%08lX)      a2 %-13s (0x%08lX)                      ",
   "      a1 %-13s (0x%08lX)      a3 %-13s (0x%08lX)                      ",

   },

*unknown_register_fmt[] = {

 "         r1 0x******    r7 0x******   r13 0x******   r19 0x******         ",
 "         r2 0x******    r8 0x******   r14 0x******   r20 0x******         ",
 "         r3 0x******    r9 0x******   r15 0x******   r21 0x******  (****) ",
 "         r4 0x******   r10 0x******   r16 0x******   r22 0x******  (ivtp) ",
 "         r5 0x******   r11 0x******   r17 0x******                        ",
 "         r6 0x******   r12 0x******   r18 0x******                        ",
 "                                                                          ",
 "      a0 ************* (0x********)      a2 ************* (0x********)    ",
 "      a1 ************* (0x********)      a3 ************* (0x********)    ",

 };


char *rprintf( int ifmt) {

	static char buf[2*SCREEN_WIDTH]; /* 2* for sprintf buf growth */

	    char num0[32], num1[32], num2[32], num3[32];



	/* Format a line of the DSP32C register display. */

	if ( 0 <= ifmt && ifmt <= 3 ) {

		sprintf( buf, register_fmt[ifmt],

			regs.cau[ifmt+ 0] & 0xFFFFFFL,
			regs.cau[ifmt+ 6] & 0xFFFFFFL,
			regs.cau[ifmt+12] & 0xFFFFFFL,
			regs.cau[ifmt+18] & 0xFFFFFFL

			);
		}

	else if ( 4 <= ifmt && ifmt <= 5 ) {

		sprintf( buf, register_fmt[ifmt],

			regs.cau[ifmt+ 0] & 0xFFFFFFL,
			regs.cau[ifmt+ 6] & 0xFFFFFFL,
			regs.cau[ifmt+12] & 0xFFFFFFL

			);
		}

	else if ( ifmt == 7) {

		strcpy( num0, DSP32C_DspfToA( &regs.dau[0].f));
		strcpy( num2, DSP32C_DspfToA( &regs.dau[2].f));

		sprintf( buf, register_fmt[ifmt],

			num0, regs.dau[0].l,
			num2, regs.dau[2].l

			);
		}

	else if ( ifmt == 8 ) {

		strcpy( num1, DSP32C_DspfToA( &regs.dau[1].f));
		strcpy( num3, DSP32C_DspfToA( &regs.dau[3].f));

		  sprintf( buf, register_fmt[ifmt],

			num1, regs.dau[1].l,
			num3, regs.dau[3].l

			);

		}

	else
		sprintf( buf, register_fmt[ifmt]);


	return( buf);


}
void registers( void) {

	int i1, j1, ni, nj, i2, color, ifmt;

	    char buffer[SCREEN_WIDTH];



	/* Display the window title bar, flags and breakpoint. */

	   i1 = r.w->ititle;
	   j1 = r.w->jtitle;
	   nj = r.w->wscroll;
        color = R_CTITLE_OFF;

	if ( execution_status == BREAKPOINTING ) {

		    sprintf( buffer, flag_fmt,

				r.w->title,

				regs.cau_flags.n ? 'N' : 'n' ,
				regs.cau_flags.z ? 'Z' : 'z' ,
				regs.cau_flags.v ? 'V' : 'v' ,
				regs.cau_flags.c ? 'C' : 'c' ,

				regs.dau_flags.n ? 'N' : 'n' ,
				regs.dau_flags.z ? 'Z' : 'z' ,
				regs.dau_flags.v ? 'V' : 'v' ,
				regs.dau_flags.u ? 'U' : 'u' ,

				breakpoint & 0xFFFFFFL

				);
		}

	else
		sprintf( buffer, unknown_flag_fmt,

			              r.w->title, breakpoint&0xFFFFFFL );


	LL_PutS( i1, j1, buffer, nj, color);




	/* Display the DSP32C registers. */

	   i1 = r.w->iscroll;
	   j1 = r.w->jscroll;

	   nj = r.w->wscroll;
	color = r.w->cscroll;


	if ( execution_status == BREAKPOINTING ) {

		for ( ifmt = 0 ; ifmt < 9 ; ifmt++ )

			LL_PutS( i1++, j1, rprintf( ifmt), nj, color);

		}

	else {

		/* Display the unknown register display. */

		for ( ifmt = 0 ; ifmt < 9 ; ifmt++ )

		    LL_PutS( i1++, j1, unknown_register_fmt[ifmt], nj, color);

		}


}






/* GENERAL COMMANDS:
 *
 *
 *
 *
 *
 */

void CommandWindowTab( void) {

	WINDOW *wnew;


	/* Toggle between the display memory and disassemble windows. */

	if ( whighlighted == &display_memory )

		wnew = &unassemble_memory;

	else
		wnew = &display_memory;



	WSelect( wnew);


}
void CommandWindowGrow( void) {


	/* Increase the size of the disassembly window by one line. */

	if ( display_memory.hscroll > 1 ) {

		display_memory.hscroll -= 1;

		  unassemble_memory.ititle  -= 1;
		  unassemble_memory.iscroll -= 1;
		  unassemble_memory.hscroll += 1;


		DataDisplay(); unassemble();

		}


}
void CommandWindowShrink( void) {


	/* Decrease the size of the disassembly window by one line. */

	if ( unassemble_memory.hscroll > 1 ) {

		display_memory.hscroll += 1;

		  unassemble_memory.ititle  += 1;
		  unassemble_memory.iscroll += 1;
		  unassemble_memory.hscroll -= 1;


		DataDisplay(); unassemble();

		}

}
void CommandWindowMaximize( void) {


	/* Maximize the size of the disassembly window. */

	display_memory.hscroll = 1;
	  
	  unassemble_memory.ititle  = 2;
	  unassemble_memory.iscroll = 3;
	  unassemble_memory.hscroll = D_HSCROLL + U_HSCROLL - 1;


	DataDisplay(); unassemble();


}
void CommandWindowMinimize( void) {


	/* Minimize the size of the disassembly window. */

	display_memory.hscroll = D_HSCROLL + U_HSCROLL - 1;
	
	  unassemble_memory.ititle  = D_HSCROLL + U_HSCROLL;
	  unassemble_memory.iscroll = D_HSCROLL + U_HSCROLL + 1;
	  unassemble_memory.hscroll = 1;


	DataDisplay(); unassemble();


}
void update_breakpoint( void) {

	int hscroll;



	/* Scroll the unassemble window to show the breakpoint. */

	hscroll = unassemble_memory.hscroll;


	if ( breakpoint < u.start )

		u.start = breakpoint;


	else if ( breakpoint >= u.start + 4*(hscroll/2) )

		u.start = breakpoint - 4*(hscroll/2);




	/* Update the display. */

	UpdateWindows();


}
void CommandGo( void) {

        void UpdateWindows( void);


	/* If already executing, tell the user they should stop first. */
	
	if ( execution_status == EXECUTING )

		PopupError( "Already executing, halt first", 0);



	/* Else if halted start continuous execution from reset. */
	
	else if ( execution_status == HALTED ) {
		

		DSP32C_Go();

		  PopupMsg( "Executing from reset until halted", 0);


		execution_status = EXECUTING;

		}



	/* Else restart continuous execution from a previous breakpoint. */

	else { /* execution_status == BREAKPOINTING */
		

		RestartFromBreakpoint();

	          PopupMsg( "Continuing execution until halted", FKEY_STEP);


		execution_status = EXECUTING;

		}




        /* Update the display. */
    
	UpdateWindows();



}
void kb_interrupt( void) {


	/* Halt if execution was broken with a kb interrupt. */
	
	DSP32C_Halt();
	
	    breakpoint = 0; execution_status = HALTED;


	/* Give the user an error message. */

	PopupError( "Execution interrupted with kbhit, DSP32C halted",

							FKEY_STEP);

	/* Update the screen. */
		
	update_breakpoint();


}
void CommandBreakpoint( unsigned long newbreak) {


	/* If already executing, tell the user they should stop first. */
	
	if ( execution_status == EXECUTING )

		PopupError( "Already executing, halt first", FKEY_STEP);



	/* Otherwise set the breakpoint and execute. */

	else { /* execution_status == HALTED or BREAKPOINTING */


	    breakpoint = newbreak;


		if ( ExecuteBreakpoint() )

			execution_status = BREAKPOINTING;
			
		else
			kb_interrupt();

	    }
   

        /* Update the screen. */
    
	update_breakpoint();


}
void CommandStep( void) {

	unsigned long newbreak;


	/* Move the breakpoint one instruction onward and execute. */

	newbreak = breakpoint + 4;
	
		CommandBreakpoint( newbreak);


}
void CommandHalt( void) {


	/* Halt the DSP32C, resetting to 0. */

	DSP32C_Halt();
	
	    breakpoint = 0; execution_status = HALTED;



	/* Give the user a message. */

	PopupMsg( "DSP32C halted", FKEY_HALT);



	/* Update the screen. */

	update_breakpoint();



}
void CommandEnableEni( void) {


	/* Enable the PIR interrupt. */

	DSP32C_SetEni( 1);


}
void CommandQuit( void) {


	/* Restore video and quit. */

	LL_VideoRestore();

	exit( 0);



}












/* DSPMON COMMAND LINE PARSING:
 *
 * These functions parse arguments specified on the DSPMON command line.
 *
 *
 *
 */

int space( void) {

	char *save;


	/* Skip over any white space in the command buffer. */

	save = cbuf;

	  if ( isspace( *cbuf) )

	      { while ( isspace( *cbuf) ) cbuf++; return( 1); }


	  else
		{ cbuf = save; return( 0); }


}
int type( int *flag) {

	struct type { int c, type_flag; } *t;

	    static struct type type_table[] = {

			{ 'b',	DSPASM_BYTE, },
			{ 'w',	DSPASM_WORD, },
			{ 'l',	DSPASM_LONG, },
			{ 'f',	DSPASM_DSPF, },
			{ 'i',	DSPASM_IEEE, },

			{   0,     0, },

			};


	/* Check for a legal type specifier. */

	for ( t = type_table ; t->c ; t++ ) {

		if ( *cbuf == t->c )

			{ cbuf++; *flag = t->type_flag; return( 1); }

		}


	return( 0);


}
int hexnum( int ndigits, unsigned long *value) {

	char *save; unsigned long num; int d, n;


	/* Check for an n digit hex number. */

	save = cbuf;

	  num = 0L; n = 0;



	  /* Skip any number of leading 0's. */

	  while ( *cbuf == '0' )

		  cbuf++;


	  /* Scan for up to ndigits significant hex digits. */

	  while ( isxdigit( *cbuf) && n < ndigits ) {

		d = *cbuf++; num <<= 4;


		  if ( '0' <= d && d <= '9' )

			num += (d - '0');

		  else if ( 'a' <= d && d <= 'f' )

			num += (d - 'a' + 0xA);

		  else if ( 'A' <= d && d <= 'F' )

			num += (d - 'A' + 0xA);

		 n++;

		}


	/* Make sure something was read in. */

	if ( cbuf > save && n <= ndigits )

		{ *value = num; return( 1); }

	else

		{ cbuf = save; return( 0); }

}
int address( unsigned long *value) {

	char *save;

	    unsigned long addr; int d, ndigits;

	        char name_buffer[MAXCOMMAND], *name; SYMBOL *s; int type;



	/* Check for a symbolic address, requiring leading space. */

	save = cbuf;

	if ( space() && (isalpha( *cbuf) || *cbuf == '_') ) {

	    name = name_buffer;

		while ( isalnum( *cbuf) || *cbuf == '_' )

			*name++ = *cbuf++;

		*name = '\0';



		/* Try to find the name on the symbol table. */

		if ( (s = SymbolicValue( name_buffer, &type)) != NULL )

			{ *value = s->value; return( 1); }


		}



	/* Check for a 6 digit hex address, again requiring leading space. */

	cbuf = save;

	if ( space() && hexnum( 6, &addr) ) {

		/* Make sure 24 bits, and round to a word location. */

		  if ( addr <= 0x00FFFFFFL )

		    { addr &= 0xFFFFFFFEL; *value = addr; return( 1); }

		}



	/* Didn't find absolute or symbolic, restore cbuf and return. */

	cbuf = save;

	    return( 0);


}
int nbytes( unsigned long *value) {

	char *save;


	/* Check for an 8 digit hex number, requiring leading space. */

	save = cbuf;

	  if ( space() && hexnum( 8, value) )

	      return( 1);

	  else
		{ cbuf = save; return( 0); }


}
char *filename( void) {

	char *save, *p;

	    static char buf[MAXCOMMAND];


	/* Require a leading space. */

	save = cbuf;

	if ( !space() )

		{ cbuf = save; return( 0); }


	/* Check for a filename. */

	p = buf;

	    while ( isprint( *cbuf) ) *p++ = *cbuf++;


	  *p = '\0';


	/* Make sure something was read in. */

	if ( p > buf )

		return( buf);

	else
		{ cbuf = save; return( NULL); }


}
int eob( void) {


	/* Check for the end of the command buffer. */

	space();

	  return( *cbuf == '\0' ? 1 : 0 );

}
int eoc( void) {


	/* Check for the end of the command buffer, with an error message. */

	space();


	  if ( *cbuf == '\0' )

		return( 1);


	  else {

		PopupError( "Illegal command", 0);

		return( 0);

		}
}








/* DSPMON COMMAND LINE:
 *
 * The following defines and functions handle the DSPMON command line.
 *
 *
 *
 *
 *
 */

#define IPROMPT		49
#define JPROMPT		0

#define ICOMMAND	49
#define JCOMMAND	2

#define CPROMPT		COLOR( INTENSE|GREY, BLACK)
#define CCOMMAND	COLOR(         GREY, BLACK)


char

    *prompt = " ";


#define CURSOR  ''



void CommandClear( void) {


	/* Clear the command line buffer. */

	cbuf = command_buffer; command_length = 0;



	/* Clear the command line display. */

	LL_PutS( IPROMPT, JPROMPT, prompt, strlen( prompt), CPROMPT);

	LL_PutS( ICOMMAND, JCOMMAND, "", MAXCOMMAND+1, CCOMMAND);
	LL_PutC( ICOMMAND, JCOMMAND, CURSOR, CCOMMAND);



}
void CommandPutc( int c) {

	int i, j;


	/* Add c to the command line buffer, catching special keys. */

	i = ICOMMAND;
	j = JCOMMAND + command_length;

	switch ( c ) {

		case BS:

			if ( command_length > 0 ) {

			    cbuf--; command_length--;

				LL_PutC( i, j--, ' ', CCOMMAND);
				LL_PutC( i, j, CURSOR, CCOMMAND);

			    }

			break;


		default:

			c &= 0xFF;

			  if ( command_length < MAXCOMMAND && isprint( c) ) {

			      *cbuf++ = c; command_length++;

				LL_PutC( i, j++,      c, CCOMMAND);
				LL_PutC( i, j  , CURSOR, CCOMMAND);

			    }

			break;


		}



}
void CommandGets( char *msg) {

	int lmsg, i, j, c;



	/* Display a message string in the command area. */

	lmsg = strlen( msg);

	    LL_PutS( ICOMMAND, JCOMMAND, msg, MAXCOMMAND+1, CCOMMAND);



	/* Get a response, storing it in the command buffer area. */

	cbuf = command_buffer; command_length = 0;

	  i = ICOMMAND;
	  j = JCOMMAND + lmsg;

	     LL_PutC( i, j, CURSOR, CCOMMAND);


  get_key:

	  switch ( c = LL_KeyIn() ) {

		case BS:

			if ( command_length > 0 ) {

			    cbuf--; command_length--;

				LL_PutC( i, j--, ' ', CCOMMAND);
				LL_PutC( i, j, CURSOR, CCOMMAND);

			    }

			goto get_key;


		case ENTER:

			break;


		default:

		   c &= 0xFF;

		     if ( isprint( c) ) {

			     if ( command_length < MAXCOMMAND-lmsg ) {

				      *cbuf++ = c; command_length++;

					LL_PutC( i, j++, c, CCOMMAND);
					LL_PutC( i, j, CURSOR, CCOMMAND);

				    }

			      goto get_key;

			    }
		}



	/* Null terminate the command buffer and return. */

	*cbuf = '\0'; cbuf = command_buffer;

}




#define SYMBOLS			2
#define ENABLE_ENI		3

#define DISPLAY_MEMORY		4
#define ENTER_MEMORY		5
#define CLEAR_MEMORY		6

#define UNASSEMBLE		7
#define LOAD			8
#define GO			9
#define SINGLE_STEP		10
#define HALT			11

#define HELP			12
#define QUIT			13

struct

  command { char *name; int flag; } command_table[] = {

	"symbols",	SYMBOLS,
	"eni",		ENABLE_ENI,

	"d",		DISPLAY_MEMORY,
	"e",		ENTER_MEMORY,
	"c",		CLEAR_MEMORY,

	"u",		UNASSEMBLE,
	"l",		LOAD,
	"g",		GO,
	"s",		SINGLE_STEP,

	"halt",		HALT,
	"h",		HALT,

	"?",		HELP,

	"quit",		QUIT,
	"q",		QUIT,

	NULL,		0,

	};



void CommandEnter( void) {

	struct command *cmd; char *p;

	    unsigned long newbreak;



	/* Process the entered command. */

	*cbuf = '\0'; cbuf = command_buffer;




	/* Look for the command on the command table. */

	space();


	   if ( *cbuf == '\0' )     /* special case no command */

	      { CommandClear(); return; }



	    for ( cmd = command_table ; cmd->name ; cmd++ ) {

		if ( StrScan( cbuf, cmd->name) )

			{ cbuf += strlen( cmd->name); break; }

		}



	/* Parse the rest of the command, and take the action. */

	switch ( cmd->flag ) {

		case DISPLAY_MEMORY:

			type( &d.type); address( &d.start);

			if ( eoc() )

			       DataDisplay();

			break;

		case ENTER_MEMORY:

			type( &e.type); address( &e.addr);

			if ( eoc() )

				enter();
			break;

		case CLEAR_MEMORY:

			address( &c.start); nbytes( &c.nbytes);

			if ( eoc() )

				clear();
			break;

		case UNASSEMBLE:

			address( &u.start);

			if ( eoc() )

				 unassemble();
			break;

		case LOAD:

			p = filename();

			if ( eoc() )

			      load( p);

			break;

		case GO:

			if ( address( &newbreak) && eoc() ) {
				
				newbreak &= 0xFFFFFFFCL; /* round to 4 byte */

				    CommandBreakpoint( newbreak);

				}

			else if ( eoc() )

				CommandGo();

			break;

		case SINGLE_STEP:

			if ( eoc() )

				CommandStep();

			break;

		case HALT:

			if ( eoc() )

				CommandHalt();
			break;

		case ENABLE_ENI:

			if ( eoc() )

				CommandEnableEni();
			break;

		case HELP:

			if ( eoc() )

				PopupHelp();
			break;

		case SYMBOLS:

			if ( eoc() )

				PopupSymbols();
			break;

		case QUIT:

			if ( eoc() )

				CommandQuit();
			break;

		default:

			PopupError( "Unknown or illegal command", 0);

			break;

		}



	/* Clear the command buffer. */
	
	CommandClear();



}
void CommandProcessor( void) {

	unsigned c;



	/* Process keyboard input. */

	getkey:


	    switch ( c = LL_KeyIn() ) {


		/* Process any single keystroke hot key commands. */

		case TAB:

			CommandWindowTab(); break;

		case UP:

			whighlighted->up(); break;

		case DOWN:

			whighlighted->down(); break;

		case PGUP:

			whighlighted->pgup(); break;

		case PGDN:

			whighlighted->pgdn(); break;

		case HOME:

			whighlighted->home(); break;

		case END:

			whighlighted->end(); break;

		case LEFT:

			CommandWindowGrow(); break;

		case RIGHT:

			CommandWindowShrink(); break;

		case CTRL_LEFT:
			
			CommandWindowMaximize(); break;
			
		case CTRL_RIGHT:
			
			CommandWindowMinimize(); break;

		case FKEY_HELP:

			PopupHelp(); break;

		case FKEY_SYMBOLS:

			PopupSymbols(); break;

		case FKEY_STEP:

			CommandStep(); break;

		case FKEY_HALT:

			CommandHalt(); break;

		case FKEY_QUIT:

			CommandQuit(); break;

		case ESC:

			CommandClear(); break;




		/* Process a completed command line. */

		case ENTER:

			CommandEnter(); break;




		/* Default to adding the character to the command buffer. */

		default:

			CommandPutc( c); break;


		}


	    goto getkey;
}









/* CHARACTER WINDOW INITIALIZATION AND UPDATING:
 *
 * Initialization for the screen and updating with new data after a command.
 *
 *
 *
 */

void InitWindows( void) {


	/* Set the video mode. */

	LL_VideoSet();



	/* Initialize the display windows. */

	CommandClear();

	      DataDisplay(); unassemble(); registers();




	/* Start off in the display window. */

	WSelect( &unassemble_memory);



}
void UpdateWindows( void) {

	/* Update all the DSPMON windows with any new data. */

	DataDisplay(); unassemble(); registers();

}












/* MAIN:
 *
 * Get DOS command line arguments and start the DSPMON command processor.
 *
 *
 *
 *
 */

#define NO_CPU			-1
#define MAXFILES		 2


void main( int argc, char **argv) {

	int nfields, ncpu, nbytes;

	char *filename[MAXFILES];

	    int ifile, nfiles, dso, dss, out;






	/* Process the command line. */

	  ncpu = NO_CPU;
	 ifile = 0;
	nfiles = 0;


	while ( --argc > 0 ) { ++argv;



		/* Check for the /ncpu DSPMUL option. */

		if ( **argv == '/' ) {


			nfields = sscanf( *argv, "/%d%n", &ncpu, &nbytes);


			if ( nfields != 1 || (*argv)[nbytes] != '\0' )

				DOS_Error( "Illegal command line option ...");


			else if ( ncpu < 0 || 3 < ncpu )

				DOS_Error( "Illegal DSPMUL cpu number ...");

			}



		/* If not an option, then assume argv is an input filename. */

		else {

			filename[ifile++] = *argv;


			    if ( ++nfiles > MAXFILES )

				DOS_Error( "Too many input files ...");
		        }

		}







	/* Initialize the DSP32C board. */

	if ( ncpu > NO_CPU )

		DSP32C_SelectCpu( ncpu);  /** for DSPMUL boards ... **/


	if ( !DSP32C_Init() )

		DOS_Warning( "Unable to intialize DSP32C board");






	/* Process any input files. */

	dso = 0;
	dss = 0;
	out = 0;

	for ( ifile = 0 ; ifile < nfiles ; ifile++ ) {


		if ( DSP32C_SrLoad( filename[ifile], NO_REPORT) )

			dso++;

		else if ( ReadSymbols( filename[ifile]) )

			dss++;

		else if ( DSP32C_WeLoad( filename[ifile], NO_REPORT) )

			out++;


		else
		      DOS_Error( "Illegal input file ...");


		}


	if ( dso > 1 || out > 1 || (dso > 0 && out > 0) )

		DOS_Error( "Too many object files on command line ...");

	else if ( dss > 1 )

		DOS_Error( "Too many symbol table files on command line ...");






	/* Initialize the DSPMON display windows. */

	InitWindows();



	/* Start the keyboard command processor. */

	CommandProcessor();


}
