/* FILE: block.dsp              Copyright (c), Symmetric Research, 1992
 *
 * This program performs a block memory move, as was discussed in the
 * Summer 1989 AT&T DSP review newsletter.
 *
 * When first looking at the DSP32C instruction set, it may not appear
 * to have a block memory move, but it does.  The way to do it is to use
 * one of the DAU tap update instructions.  When placed inside a do loop,
 * this will move 32 bit words without instruction fetching.
 * 
 * Using a tap update instruction to do the move is preferable to passing
 * the 32 bit data through an accumulator.  When data is loaded into the
 * the accumulator, it is checked to see if it is a "dirty" zero, i.e. a
 * floating point number with a zero exponent and a nonzero mantissa.  If
 * it is, then the data is altered and changed to a "clean" zero, with all
 * bits zero.
 * 
 * For a block memory move of arbitrary data this is not acceptable.  The
 * tap update part of a DAU instruction fortunately does no such checks on
 * the data and can move blocks efficiently.
 *
 * To run the program below, assemble it, and then run it by hand in the
 * DSPMON monitor, checking the x and y arrays arrays before and after
 * running.
 *
 * Finally, note that the target array could be one of the parallel ports
 * on the board.  This would be useful when driving a D/A converter connected
 * to the port.
 *
 *
 */

constant

        RESET           = 0,
        DATA            = 100,
        N               = 4;


<DATA>

    long xsrc[] = { 1, 2, 3, 4 }, ydest[N];



<RESET> main {

        register

                x       = r1,
                y       = r2;


        /* Point to the x and y arrays and do the block move. */

        x = xsrc;
        y = ydest;

                do N { a0 = (*y++ = *x++) + a0; }


        /* Wait for the PC. */

        wait: goto wait;

}
